import { VNode } from 'vue';
import { NormalizedScopedSlot } from 'vue/types/vnode';

type DefineComponent<
  P = NonNullable<unknown>,
  S = NonNullable<unknown>,
  E = NonNullable<unknown>,
  M = NonNullable<unknown>,
> = {
  new (): {
    $props: P
    $slots: S
    $emit: E
  } & M
}

type ObjectEmitsOptions = {
  [event: string]: (...args: any[]) => void | boolean
}

declare type UnionToIntersection<U> = (U extends any ? (k: U) => void : never) extends (
  k: infer I,
) => void
  ? I
  : never

declare type EmitFn<
  Options = ObjectEmitsOptions,
  Event extends keyof Options = keyof Options,
> =
  Options extends Array<infer V>
    ? (event: V, ...args: any[]) => void
    : NonNullable<unknown> extends Options
      ? (event: string, ...args: any[]) => void
      : UnionToIntersection<
          {
            [key in Event]: Options[key] extends (...args: infer Args) => any
              ? (event: key, ...args: Args) => void
              : (event: key, ...args: any[]) => void
          }[Event]
        >

// eslint-disable-next-line import/no-unresolved


/**
 * Columns
 */
type DTTextAlign = 'center' | 'left' | 'right'

type DTColumnProps = {
  readonly field: string
  readonly title: string
  readonly orderable: boolean
  readonly searchable: boolean
  readonly width: string | undefined
  readonly textAlign: DTTextAlign | undefined
}

type DTColumnParams = {
  readonly field: string
  readonly title: string | null
  readonly orderable: boolean
  readonly searchable: boolean
}
type DTColumnSlots = {
  readonly cell: NormalizedScopedSlot | undefined
  readonly filter: NormalizedScopedSlot | undefined
}
type DTColumnAppearance = {
  readonly styleObject: Record<string, string | number>
  readonly classObject: Record<string, boolean>
}

type DTColumn = {
  readonly index: number
  readonly params: DTColumnParams
  readonly slots: DTColumnSlots
  readonly appearance: DTColumnAppearance
}

/**
 * Table data
 */
type DTSource = 'server' | 'client'

type DTRowItemValue = string | number | boolean | null
type DTRowItem = Record<string, DTRowItemValue>
type DTRow = {
  readonly index: number
  readonly number: number
  readonly item: DTRowItem
}

type DTTableData = {
  readonly total: number
  readonly filtered: number
  readonly rows: DTRow[]
}

/**
 * Order
 */
type DTOrderDirection = 'asc' | 'desc'
type DTOrderColumn = {
  readonly column: string
  readonly direction: DTOrderDirection
}
type DTOrder = DTOrderColumn

/**
 * Error
 */
type DTError = {
  message: string
  description?: string | null
}

interface DataTableProps {
  // data
  source?: DTSource
  url?: string | null
  items?: DTRowItem[]

  // pagination
  pagination?: boolean
  rowsPerPageCount?: number
  rowsPerPageOptions?: number[]

  // search
  searching?: boolean

  // order
  orderBy?: string | null
  orderDirection?: DTOrderDirection

  // row-selection
  rowSelection?: boolean

  // actions
  actions?: boolean

  // numbering
  numbering?: boolean

  // scrollX
  scrollX?: boolean

  // fixed columns
  fixedColumnsStart?: number
  fixedColumnsEnd?: number

  // rows clickable
  rowsClickable?: boolean
}

interface DataTableSlots {
  topLeftBeforeActions(): VNode[]

  topLeftAfterActions(): VNode[]

  topRight(): VNode[]

  actions(scope: { row: DTRow }): VNode
}

interface DataTableEmitsOptions {
  /**
   * Emitted when the selected rows array changes.
   * @param {DTRow[]} value - Selected rows.
   */
  'update:selected-rows'(value: DTRow[]): void

  /**
   * Emitted when the row is clicked.
   * @param {DTRow} value - Clicked row.
   */
  'row-click'(value: DTRow): void
}

declare type DataTableEmits = EmitFn<DataTableEmitsOptions>

interface DataTableMethods {
  /**
   * Reload table data.
   */
  reload(): Promise<void>
}

declare const DataTable: DefineComponent<
  DataTableProps,
  DataTableSlots,
  DataTableEmits,
  DataTableMethods
>

interface DataTableColumnProps {
  title: string
  orderable?: boolean
  searchable?: boolean
  width?: string
  textAlign?: DTTextAlign
}

interface DataTableColumnSlots {
  cell(scope: { row: DTRow }): VNode
}

declare const DataTableColumn: DefineComponent<DataTableColumnProps, DataTableColumnSlots>

export { type DTColumn, type DTColumnProps, type DTError, type DTOrder, type DTOrderColumn, type DTOrderDirection, type DTRow, type DTRowItem, type DTRowItemValue, type DTSource, type DTTableData, type DTTextAlign, DataTable, DataTableColumn, type DataTableEmits, type DataTableEmitsOptions, type DataTableMethods, type DataTableProps, type DataTableSlots };
